@@ -757,35 +757,20 @@ func (f *file) messageToSymbolsHelper(msg ir.MessageValue, index int, parents []
 // resolveASTDefinition is a helper for resolving the [ast.DeclDef] to the *[symbol], if
 // there is a matching indexed *[symbol].
 func (f *file) resolveASTDefinition(def ast.DeclDef, defName ir.FullName) *symbol {
-	// No workspace, we cannot resolve the AST definition
+	// First check if the definition is in the current file.
+	if def.Span().Path() == f.file.Path() {
+		return f.referenceableSymbols[defName]
+	}
+	// No workspace, we cannot resolve the AST definition from outside of the file.
 	if f.workspace == nil {
 		return nil
 	}
-	// We resolve the import path of the span of the AST definition and search for it in our
-	// workspace.
-	fileInfo, ok := f.workspace.fileNameToFileInfo[def.Span().Path()]
-	if !ok {
-		// Unable to resolve an importable path for the file in our workspace, log a debug
-		// statement and return no definition.
-		f.lsp.logger.Debug(
-			"unable to resolve an importable file path for local path",
-			slog.String("file", f.uri.Filename()),
-			slog.String("local path", def.Span().Path()),
-		)
-		return nil
-	}
-	file, ok := f.workspace.PathToFile()[fileInfo.Path()]
-	if !ok {
-		// Check current file
-		if def.Span().Path() != f.objectInfo.Path() {
-			// If there is no file for the [ast.DeclDef] span, which can if it is a predeclared
-			// type, for example, or if the file we are checking has not indexed imports, then
-			// we return nil.
-			return nil
+	for _, file := range f.workspace.PathToFile() {
+		if file.file.Path() == def.Span().Path() {
+			return file.referenceableSymbols[defName]
 		}
-		file = f
 	}
-	return file.referenceableSymbols[defName]
+	return nil
 }
 
 // SymbolAt finds a symbol in this file at the given cursor position, if one exists.
