@@ -19,10 +19,10 @@ import (
 	"errors"
 	"fmt"
 	"runtime/debug"
-	"slices"
 	"strings"
 	"unicode/utf16"
 
+	"buf.build/go/standard/xslices"
 	"github.com/bufbuild/buf/private/buf/bufformat"
 	"github.com/bufbuild/protocompile/experimental/ir"
 	"github.com/bufbuild/protocompile/parser"
@@ -32,8 +32,13 @@ import (
 
 const (
 	serverName = "buf-lsp"
+
+	maxSymbolResults = 1000
 )
 
+// The subset of SemanticTokenTypes that we support.
+// Must match the order of [semanticTypeLegend].
+// https://microsoft.github.io/language-server-protocol/specifications/lsp/3.18/specification/#textDocument_semanticTokens
 const (
 	semanticTypeType = iota
 	semanticTypeStruct
@@ -43,15 +48,35 @@ const (
 	semanticTypeInterface
 	semanticTypeMethod
 	semanticTypeDecorator
+	semanticTypeNamespace
+)
+
+// The subset of SemanticTokenModifiers that we support.
+// Must match the order of [semanticModifierLegend].
+// Semantic modifiers are encoded as a bitset, hence the shifted iota.
+// https://microsoft.github.io/language-server-protocol/specifications/lsp/3.18/specification/#textDocument_semanticTokens
+const (
+	semanticModifierDeprecated = 1 << iota
+	semanticModifierDefaultLibrary
 )
 
 var (
-	// These slices must match the order of the indices in the above const block.
+	// These slices must match the order of the indices in the above const blocks.
 	semanticTypeLegend = []string{
-		"type", "struct", "variable", "enum",
-		"enumMember", "interface", "method", "decorator",
+		"type",
+		"struct",
+		"variable",
+		"enum",
+		"enumMember",
+		"interface",
+		"method",
+		"decorator",
+		"namespace",
+	}
+	semanticModifierLegend = []string{
+		"deprecated",
+		"defaultLibrary", // maps to builtin values
 	}
-	semanticModifierLegend = []string{}
 )
 
 // server is an implementation of [protocol.Server].
@@ -100,8 +125,6 @@ func (s *server) Initialize(
 		TokenModifiers []string `json:"tokenModifiers"`
 	}
 	type SemanticTokensOptions struct {
-		protocol.WorkDoneProgressOptions
-
 		Legend SemanticTokensLegend `json:"legend"`
 		Full   bool                 `json:"full"`
 	}
@@ -122,18 +145,14 @@ func (s *server) Initialize(
 			},
 			CompletionProvider: &protocol.CompletionOptions{
 				ResolveProvider:   true,
-				TriggerCharacters: []string{" ", ".", "("},
-			},
-			DefinitionProvider: &protocol.DefinitionOptions{
-				WorkDoneProgressOptions: protocol.WorkDoneProgressOptions{WorkDoneProgress: true},
+				TriggerCharacters: []string{".", "\"", "/"},
 			},
+			DefinitionProvider:         &protocol.DefinitionOptions{},
+			TypeDefinitionProvider:     &protocol.TypeDefinitionOptions{},
 			DocumentFormattingProvider: true,
 			HoverProvider:              true,
-			ReferencesProvider: &protocol.ReferenceOptions{
-				WorkDoneProgressOptions: protocol.WorkDoneProgressOptions{WorkDoneProgress: true},
-			},
+			ReferencesProvider:         &protocol.ReferenceOptions{},
 			SemanticTokensProvider: &SemanticTokensOptions{
-				WorkDoneProgressOptions: protocol.WorkDoneProgressOptions{WorkDoneProgress: true},
 				Legend: SemanticTokensLegend{
 					TokenTypes:     semanticTypeLegend,
 					TokenModifiers: semanticModifierLegend,
@@ -141,6 +160,7 @@ func (s *server) Initialize(
 				Full: true,
 			},
 			WorkspaceSymbolProvider: true,
+			DocumentSymbolProvider:  true,
 		},
 		ServerInfo: info,
 	}, nil
@@ -152,20 +172,7 @@ func (s *server) Initialized(
 	ctx context.Context,
 	params *protocol.InitializedParams,
 ) error {
-	workspaceCapabilities := s.initParams.Load().Capabilities.Workspace
-	if workspaceCapabilities == nil {
-		return nil
-	}
-	didChangeConfiguration := workspaceCapabilities.DidChangeConfiguration
-	if didChangeConfiguration != nil && didChangeConfiguration.DynamicRegistration {
-		// The error is logged for us by the client wrapper.
-		_ = s.client.RegisterCapability(ctx, &protocol.RegistrationParams{
-			Registrations: []protocol.Registration{
-				{Method: protocol.MethodWorkspaceDidChangeConfiguration},
-			},
-		})
-	}
-
+	// No initialization required.
 	return nil
 }
 
@@ -196,23 +203,6 @@ func (s *server) Exit(ctx context.Context) error {
 	return s.lsp.conn.Close()
 }
 
-// DidChangeConfiguration is sent whenever the client changes its config settings.
-func (s *server) DidChangeConfiguration(
-	ctx context.Context,
-	params *protocol.DidChangeConfigurationParams,
-) error {
-	// We need to refresh every open file's settings, and refresh the file
-	// itself.
-	s.fileManager.uriToFile.Range(func(_ protocol.URI, file *file) bool {
-		if file.IsOpenInEditor() {
-			file.RefreshSettings(ctx)
-			file.Refresh(ctx)
-		}
-		return true
-	})
-	return nil
-}
-
 // -- File synchronization methods.
 
 // DidOpen is called whenever the client opens a document. This is our signal to parse
@@ -222,9 +212,8 @@ func (s *server) DidOpen(
 	params *protocol.DidOpenTextDocumentParams,
 ) error {
 	file := s.fileManager.Track(params.TextDocument.URI)
-	file.RefreshSettings(ctx)
+	file.RefreshWorkspace(ctx)
 	file.Update(ctx, params.TextDocument.Version, params.TextDocument.Text)
-	file.Refresh(ctx)
 	return nil
 }
 
@@ -241,7 +230,6 @@ func (s *server) DidChange(
 	}
 
 	file.Update(ctx, params.TextDocument.Version, params.ContentChanges[0].Text)
-	file.Refresh(ctx)
 	return nil
 }
 
@@ -257,7 +245,7 @@ func (s *server) DidSave(
 		// Update for a file we don't know about? Seems bad!
 		return fmt.Errorf("received update for file that was not open: %q", params.TextDocument.URI)
 	}
-	file.Refresh(ctx)
+	file.RefreshWorkspace(ctx)
 	return nil
 }
 
@@ -335,7 +323,9 @@ func (s *server) DidClose(
 	ctx context.Context,
 	params *protocol.DidCloseTextDocumentParams,
 ) error {
-	s.fileManager.Close(ctx, params.TextDocument.URI)
+	if file := s.fileManager.Get(params.TextDocument.URI); file != nil {
+		file.Close(ctx)
+	}
 	return nil
 }
 
@@ -346,17 +336,12 @@ func (s *server) Hover(
 	ctx context.Context,
 	params *protocol.HoverParams,
 ) (*protocol.Hover, error) {
-	file := s.fileManager.Get(params.TextDocument.URI)
-	if file == nil {
-		return nil, nil
-	}
-
-	symbol := file.SymbolAt(ctx, params.Position)
+	symbol := s.getSymbol(ctx, params.TextDocument.URI, params.Position)
 	if symbol == nil {
 		return nil, nil
 	}
 
-	docs := symbol.FormatDocs(ctx)
+	docs := symbol.FormatDocs()
 	if docs == "" {
 		return nil, nil
 	}
@@ -380,43 +365,46 @@ func (s *server) Definition(
 	ctx context.Context,
 	params *protocol.DefinitionParams,
 ) ([]protocol.Location, error) {
-	file := s.fileManager.Get(params.TextDocument.URI)
-	if file == nil {
-		return nil, nil
-	}
-
-	progress := newProgressFromClient(s.lsp, &params.WorkDoneProgressParams)
-	progress.Begin(ctx, "Searching")
-	defer progress.Done(ctx)
-
-	symbol := file.SymbolAt(ctx, params.Position)
+	symbol := s.getSymbol(ctx, params.TextDocument.URI, params.Position)
 	if symbol == nil {
 		return nil, nil
 	}
-
 	return []protocol.Location{
 		symbol.Definition(),
 	}, nil
 }
 
+// TypeDefinition is the entry point for go-to type-definition.
+func (s *server) TypeDefinition(
+	ctx context.Context,
+	params *protocol.TypeDefinitionParams,
+) ([]protocol.Location, error) {
+	symbol := s.getSymbol(ctx, params.TextDocument.URI, params.Position)
+	if symbol == nil {
+		return nil, nil
+	}
+	return []protocol.Location{
+		symbol.TypeDefinition(),
+	}, nil
+}
+
 // References is the entry point for get-all-references.
 func (s *server) References(
 	ctx context.Context,
 	params *protocol.ReferenceParams,
 ) ([]protocol.Location, error) {
-	file := s.fileManager.Get(params.TextDocument.URI)
-	if file == nil {
-		return nil, nil
-	}
-	progress := newProgressFromClient(s.lsp, &params.WorkDoneProgressParams)
-	progress.Begin(ctx, "Searching")
-	defer progress.Done(ctx)
-
-	symbol := file.SymbolAt(ctx, params.Position)
+	symbol := s.getSymbol(ctx, params.TextDocument.URI, params.Position)
 	if symbol == nil {
 		return nil, nil
 	}
-	return symbol.References(), nil
+	// We deduplicate the references here in the case where a file's symbols have not yet
+	// been refreshed, but a new file with references to symbols in said file is opened. This
+	// can cause duplicate references to be appended and not all clients deduplicate the
+	// returned references.
+	//
+	// We also do not want to refresh all symbols in the workspace when a single file is
+	// interacted with, since that could be detrimental to performance.
+	return xslices.Deduplicate(symbol.References(params.Context.IncludeDeclaration)), nil
 }
 
 // Completion is the entry point for code completion.
@@ -429,6 +417,9 @@ func (s *server) Completion(
 		return nil, nil
 	}
 	items := getCompletionItems(ctx, file, params.Position)
+	if len(items) == 0 {
+		return nil, nil
+	}
 	return &protocol.CompletionList{Items: items}, nil
 }
 
@@ -449,9 +440,6 @@ func (s *server) SemanticTokensFull(
 	if file == nil {
 		return nil, nil
 	}
-	progress := newProgressFromClient(s.lsp, &params.WorkDoneProgressParams)
-	progress.Begin(ctx, "Processing Tokens")
-	defer progress.Done(ctx)
 	// In the case where there are no symbols for the file, we return nil for SemanticTokensFull.
 	// This is based on the specification for the method textDocument/semanticTokens/full,
 	// the expected response is the union type `SemanticTokens | null`.
@@ -465,13 +453,16 @@ func (s *server) SemanticTokensFull(
 		encoded           []uint32
 		prevLine, prevCol uint32
 	)
-	for i, symbol := range file.symbols {
-		progress.Report(ctx, fmt.Sprintf("%d/%d", i+1, len(file.symbols)), float64(i)/float64(len(file.symbols)))
+	for _, symbol := range file.symbols {
 		var semanticType uint32
-		if symbol.isOption {
+		var semanticModifier uint32
+		_, ok := symbol.kind.(*option)
+		if ok {
 			semanticType = semanticTypeDecorator
 		} else {
 			switch symbol.ir.Kind() {
+			case ir.SymbolKindPackage:
+				semanticType = semanticTypeNamespace
 			case ir.SymbolKindMessage:
 				semanticType = semanticTypeStruct
 			case ir.SymbolKindEnum:
@@ -480,6 +471,7 @@ func (s *server) SemanticTokensFull(
 				// For predeclared types, we set semanticType to semanticTypeType
 				if symbol.IsBuiltIn() {
 					semanticType = semanticTypeType
+					semanticModifier += semanticModifierDefaultLibrary
 				} else {
 					semanticType = semanticTypeVariable
 				}
@@ -495,6 +487,9 @@ func (s *server) SemanticTokensFull(
 				continue
 			}
 		}
+		if _, ok := symbol.ir.Deprecated().AsBool(); ok {
+			semanticModifier += semanticModifierDeprecated
+		}
 
 		startLocation := symbol.span.Location(symbol.span.Start, positionalEncoding)
 		endLocation := symbol.span.Location(symbol.span.End, positionalEncoding)
@@ -512,7 +507,7 @@ func (s *server) SemanticTokensFull(
 			if i == startLocation.Line {
 				symbolLen -= uint32(startLocation.Column - 1)
 			}
-			encoded = append(encoded, newLine-prevLine, newCol, symbolLen, semanticType, 0)
+			encoded = append(encoded, newLine-prevLine, newCol, symbolLen, semanticType, semanticModifier)
 			prevLine = newLine
 			if i == startLocation.Line {
 				prevCol = uint32(startLocation.Column - 1)
@@ -521,6 +516,9 @@ func (s *server) SemanticTokensFull(
 			}
 		}
 	}
+	if len(encoded) == 0 {
+		return nil, nil
+	}
 	return &protocol.SemanticTokens{Data: encoded}, nil
 }
 
@@ -529,45 +527,48 @@ func (s *server) Symbols(
 	ctx context.Context,
 	params *protocol.WorkspaceSymbolParams,
 ) ([]protocol.SymbolInformation, error) {
-	const maxResults = 1000 // Limit results to avoid overwhelming clients
 	query := strings.ToLower(params.Query)
-
 	var results []protocol.SymbolInformation
 	for _, file := range s.fileManager.uriToFile.Range {
-		if file.ir.IsZero() {
-			continue
-		}
-		// Search through all symbols in this file.
-		for _, sym := range file.symbols {
-			if sym.ir.IsZero() {
-				continue
-			}
-			// Only include definitions: static and referenceable symbols.
-			// Skip references, imports, builtins, and tags
-			_, isStatic := sym.kind.(*static)
-			_, isReferenceable := sym.kind.(*referenceable)
-			if !isStatic && !isReferenceable {
-				continue
-			}
-			symbolInfo := sym.GetSymbolInformation()
-			if symbolInfo.Name == "" {
-				continue // Symbol information not supported for this symbol.
-			}
-			// Filter by query (case-insensitive substring match)
-			if query != "" && !strings.Contains(strings.ToLower(symbolInfo.Name), query) {
-				continue
-			}
-			results = append(results, symbolInfo)
-			if len(results) >= maxResults {
+		for symbol := range file.GetSymbols(query) {
+			results = append(results, symbol)
+			if len(results) > maxSymbolResults {
 				break
 			}
 		}
 	}
-	slices.SortFunc(results, func(a, b protocol.SymbolInformation) int {
-		if a.Name != b.Name {
-			return strings.Compare(a.Name, b.Name)
-		}
-		return strings.Compare(a.ContainerName, b.ContainerName)
-	})
 	return results, nil
 }
+
+// DocumentSymbol is the entry point for document symbol search.
+func (s *server) DocumentSymbol(ctx context.Context, params *protocol.DocumentSymbolParams) (
+	result []any, // []protocol.SymbolInformation
+	err error,
+) {
+	file := s.fileManager.Get(params.TextDocument.URI)
+	if file == nil {
+		return nil, nil
+	}
+	anyResults := []any{}
+	for symbol := range file.GetSymbols("") {
+		anyResults = append(anyResults, symbol)
+		if len(anyResults) > maxSymbolResults {
+			break
+		}
+	}
+	return anyResults, nil
+}
+
+// getSymbol is a helper function that gets the *[symbol] for the given [protocol.URI] and
+// [protocol.Position].
+func (s *server) getSymbol(
+	ctx context.Context,
+	uri protocol.URI,
+	position protocol.Position,
+) *symbol {
+	file := s.fileManager.Get(uri)
+	if file == nil {
+		return nil
+	}
+	return file.SymbolAt(ctx, position)
+}
